///////////////////////////////////////////////////////////////////////////////////////////////////
//
//	Source code for BOLT-PIC18F2550 to read GPS navigation information using Skylab SKM53 device.
//
//		http://www.puntoflotante.net/
//
//	Microcontroller: PIC18F2550
//	Compiler: MPLAB IDE V.8.63 and C18 V3.40
//	Moises Melendez, sesowm@hotmail.com
//	June 2015
//	Program that shows the following navigation variables:
//
//		Latitude:	ddmmss.s K
//		Longitude:  dddmmss.s P
//		Altitude:	in meters
//		Velocity:	in km/h
//
//	 Where
//		dd and ddd are sexagesimal degrees
//		mm are minutes
//		ss.s are seconds
//		K is one of the next values: 
//			N is for north
//			S is for south
//		P is one of the next values: 
//			E is for east
//			W is for west
//
//	This navigation information is send to hardware serial port and LCD in the BOLT-PIC18F2550 
//	microcontroller system.
//	
// 	The BOLT-PIC18F2550 systems reads SW2 and SW1 to determine the navigation variable  to be sent 
//	to LCD according to the following table:
//		_______________________________
//
//		SW2	SW1		Navigation variable
//		________	___________________
//		 0   0			Latitude
//		 0   1			Longitude
//		 1   0			Altitude
//		 1   1			Velocity
//		_______________________________
//
//	Information on the LCD is as follows:
//
//		For latitude 	 	Lat:ddmm'ss.s"K
//		For longitude 		Lo:dddmm'ss.s"P
// 		For altitude		 Alt: nnnnnn m
//		For velocity 		Vel: vvvvv  km/h
//
//	If navigation data is not accurate, the system sends a string like this to LCD screen:
//
//		For latitude 	 	Lat: unavailable
//		For longitude 		Lon: unavailable
// 		For altitude		Alt: unavailable
//		For velocity 		Vel: unavailable
//
//	If data from SKM53 sensor is corrupted, the system sends the next message to LCD screen:
//	
//		GPS unavailable!
//
//	The program sends navigation data to serial port with the next format:
//
//		Latitude: ddmm'ss.s" K   Longitude: ddmm'ss.s" P   Altitude: nnnnnn m   Velocity: vvvvv km/h 
//
//	If navigation data is not accurate, the system sends a string like this:
//
//		Latitude: XXXXXXXXXXX X   Longitude: XXXXXXXXXXX X   Altitude: XXXXXX m   Velocity: XXXXX km/h
//
//	If data from SKM53 sensor is corrupted, the system sends the next message to serial port:
//
//		Unable to access to GPS information!
//
//	Data can be displayed with the following frequency:
//		1 second
//		2 seconds
//		5 seconds
//		10 seconds
//
//	For that, line 147 should be amended, uncomment only the required line of code.
//
//	The information is obtained from the navigation sensor Skylab SKM53 by the  BOLT-PIC18F2550 
//	microcontroller system.
//		
//		The connections are as follows:			
//									   
//				 	
//	    SKM53 
//	  connector									J6
//				  _______________
//				 |				 |
//				_|3.3 V regulator|_
//			   / |_______________| \		
//		VCC___/	 		  			|			RA4	(SW1)
//							 		|			
//		GND________________ 		|			RA5	(SW2)						BOLT-PIC18F2550 								
//					 	    \		|			
//		NC			________|_______|___________RC0	Software RX_pic			microcontroller system
//				   /		|		|	
//		RST		  /	   _____|_______|___________RC1	Software TX_pic
//				 /	  /		|		|
//		TXD_____/    / 	  	|		 \__________+5V
//					/		 \
//		RXD________/		  \_________________GND
//
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <p18cxxx.h>
#include <delays.h>
#include <string.h>
#include <stdio.h>		// sprintf() library
#include <stdlib.h>		// atoi(),atof() library
#include <usart.h>
#include <sw_uart.h>
#include "18F2550BOLT.h"
#include "PUERTO-SERIAL-BOLT.h"

#define BAUD_factor 	14400
#define CLK_factor 		72000000
#define INCOMING_SIZE_1	84
#define INCOMING_SIZE_2	84
#define CONV			const far rom char*
#define CR()			putcUSART(13);delay_ms(2)
#define ERROR			0
#define OK				1

#define TEST	"$PMTK000*32"					//Test packet
#define HOT	 	"$PMTK101*32"					//Hot restart
#define STAND	"$PMTK161,0*28"					//Leave and enter standby mode by PMTK command
#define WAKEUP	"$PMTK161,1*29"					//wake up GPS receiver
#define	RATE1S	"$PMTK300,1000,0,0,0,0*1C"		//1 second
#define	RATE2S	"$PMTK300,2000,0,0,0,0*1F"		//2 seconds
#define	RATE5S	"$PMTK300,5000,0,0,0,0*18"		//5 seconds	
#define	RATE10S	"$PMTK300,10000,0,0,0,0*2C"		//10 seconds
#define	DATUM_I "$PMTK330,0*2E"					//WGS1984, International
#define	DATUM_1 "$PMTK330,138*24"	 			//North American 1927,	Mexico
#define	DATUM_2 "$PMTK330,144*2F"				//North American 1983,	Mexico, Central America
#define GGA		"$PMTK314,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"	//Global positioning system fixed data
#define GLL		"$PMTK314,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"	//Geographic position, latitude, longitude
#define RMC		"$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"	//Recomended minimum specific GNSS data
#define VTG		"$PMTK314,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"	//Course over ground and ground speed
#define GGA_VTG "$PMTK314,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28"	//GGA and VTG
#define DEFAULT	"$PMTK314,-1*04"		 							//Restores the system default setting

//******************Frequency*******************
//#define	FREQUENCY RATE1S		//1 second
//#define	FREQUENCY RATE2S		//2 seconds
//#define	FREQUENCY RATE5S		//5 seconds	
  #define	FREQUENCY RATE10S		//10 seconds
//**********************************************

extern void _startup( void ); // See c018i.c in your C18 compiler dir 
#pragma code _RESET_INTERRUPT_VECTOR = 0x000800 
void _reset( void ) 
{ 
    _asm goto _startup _endasm 
} 

#pragma code

void sendCommand2GPS(const far rom char * command);			// Function to send commands to SKM53
void initConfig(void);										// Initial configurations
void convert2hex(char* csumArray, unsigned char data);		// Converts data to hexadecimal ascii string
void write2HWserie(const far rom char* message);			// Writes a string in rom to serial port
int checksum(char * serialData, char * checksumData);		// XOR checksum
void flush(void);											// flush input data buffers
void writeData2HWserie(char *data);							// Writes data to hardware serial port
void susCharX(char * splitArray, int nChars);				// Replaces the elements of a string with X characters
void GPSprocess(void);										// Function that process incoming data
void writeLCDmsg(const far rom char * message);				// Writes a rom string to LCD screen
void writeData2LCD(char * data);							// Writes a ram string to LCD screen
void writeLCDmsgRaw(const far rom char * mess);				// Writes a raw string to LCD screen
void DelayTXBitUART(void);									//
void DelayRXHalfBitUART(void);								// Configuration for software serial port
void DelayRXBitUART(void);									//
 
char serialIncomingData_1[INCOMING_SIZE_1];			// Input serila buffers
char serialIncomingData_2[INCOMING_SIZE_2];

void main()
{
	GPSprocess();		// One function 
} 

void GPSprocess(void)
{
	int i, j, initIndex=0;			//Index for serial buffer input
	int Integer,Decimal;
	int STATUS;						//Variable for conection control
	char checksum_s1[3]={0,0,0};	//checksum from GPS string 1
	char checksum_s2[3]={0,0,0};	//checksum from GPS string 2
	char latitude[11];
	char N_S[3];
	char longitude[12];
	char E_W[3];
	char fixIndicator[3];
	char altitude[8];
	char reference[3];
	char speedGround[11];
	char mode[4];
	char latitudeDeg[12];
	char longitudeDeg[13];
	char decimal[7];
	unsigned char input;
	float decimalPart;
	initConfig();					//Initial configurations
	while(1)
	{
		flush();
		initIndex=0;
		do{	
			serialIncomingData_1[initIndex]=getcUART();
			initIndex++;
		}while(serialIncomingData_1[initIndex-1]!='*');
		getsUART(checksum_s1,2);
		initIndex=0;
		do{	
			serialIncomingData_2[initIndex]=getcUART();
			initIndex++;
		}while(serialIncomingData_2[initIndex-1]!='*');
		getsUART(checksum_s2,2);							// Reads software serial data 

		STATUS=checksum(serialIncomingData_1,checksum_s1);	// If the data are reliable

		if(STATUS==OK)
		{
			i=0;
			do{	i++;
			}while(serialIncomingData_1[i-1]!=',');	//1					
			j=0;
			do{ i++;
			}while(serialIncomingData_1[i-1]!=',');	//2	
			j=0;
			do{ latitude[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			latitude[j]=0;							//3	Latitude
			j=0;
			do{	N_S[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			N_S[j]=0;								//4 N/S
			j=0;
			do{	longitude[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			longitude[j]=0;							//5 Longitude
			j=0;
			do{	E_W[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			E_W[j]=0;								//6	E/W			
			j=0;
			do{	fixIndicator[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			fixIndicator[j]=0;						//7	Position Fix Indicator	
			do{	i++;
			}while(serialIncomingData_1[i-1]!=',');	//8
			do{	i++;
			}while(serialIncomingData_1[i-1]!=',');	//9
			j=0;
			do{	altitude[j]=serialIncomingData_1[i];
				i++;j++;
			}while(serialIncomingData_1[i-1]!=',');	
			altitude[j]=0;							//10 Altitude
			i=0;
			do{	i++;
			}while(serialIncomingData_2[i-1]!=',');	//1				
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//2	
			j=0;
			do{ reference[j]=serialIncomingData_2[i];
				i++;j++;
			}while(serialIncomingData_2[i-1]!=',');	
			reference[j]=0;							//3	Reference
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//4
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//5
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//6	
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//7	
			j=0;
			do{ speedGround[j]=serialIncomingData_2[i];
				i++;j++;
			}while(serialIncomingData_2[i-1]!=',');	//8	Speed over ground
			speedGround[j]=0;
			j=0;
			do{ i++;
			}while(serialIncomingData_2[i-1]!=',');	//9	
			j=0;
			do{ mode[j]=serialIncomingData_2[i];
				i++;j++;
			}while(serialIncomingData_2[i-1]!=0);	//10 Mode
			mode[j]=0;
			if(mode[0]!='N')
				speedGround[strlen(speedGround)-1]=0;	
			else
				susCharX(speedGround,4);
			if(fixIndicator[0]!='0')
			{
				latitude[strlen(latitude)-1]=0;
				N_S[strlen(N_S)-1]=0;
				longitude[strlen(longitude)-1]=0;
				E_W[strlen(E_W)-1]=0;
				altitude[strlen(altitude)-1]=0;
				decimal[0]='0';
				decimal[1]=latitude[4];
				decimal[2]=latitude[5];
				decimal[3]=latitude[6];
				decimal[4]=latitude[7];
				decimal[5]=latitude[8];
				decimal[6]=NULL;
				decimalPart = 60.0 * atof(decimal);
				for(i=0; i<7;i++) decimal[i]=0;
				Integer=decimalPart;
				Decimal=(decimalPart-Integer)*10000; 
				sprintf(decimal,(CONV)"%02d.%01d",Integer,Decimal);	// Float number to string
				latitudeDeg[0]=latitude[0];
				latitudeDeg[1]=latitude[1];
				latitudeDeg[2]='';
				latitudeDeg[3]=latitude[2];
				latitudeDeg[4]=latitude[3];
				latitudeDeg[5]='\'';
				latitudeDeg[6]=decimal[0];
				latitudeDeg[7]=decimal[1];
				latitudeDeg[8]=decimal[2];
				latitudeDeg[9]=decimal[3];
				latitudeDeg[10]='\"';
				latitudeDeg[11]=NULL;				
				decimal[0]='0';
				decimal[1]=longitude[5];
				decimal[2]=longitude[6];
				decimal[3]=longitude[7];
				decimal[4]=longitude[8];
				decimal[5]=longitude[9];		// Transforms the data stream into a suitable format 
				decimal[6]=NULL;
				decimalPart = 60.0 * atof(decimal);
				for(i=0; i<7;i++) decimal[i]=0;
				Integer=decimalPart;
				Decimal=(decimalPart-Integer)*10000; 
				sprintf(decimal,(CONV)"%02d.%01d",Integer,Decimal);
				longitudeDeg[0]=longitude[0];
				longitudeDeg[1]=longitude[1];
				longitudeDeg[2]=longitude[2];
				longitudeDeg[3]='';
				longitudeDeg[4]=longitude[3];
				longitudeDeg[5]=longitude[4];
				longitudeDeg[6]='\'';
				longitudeDeg[7]=decimal[0];
				longitudeDeg[8]=decimal[1];
				longitudeDeg[9]=decimal[2];
				longitudeDeg[10]=decimal[3];
				longitudeDeg[11]='\"';
				longitudeDeg[12]=NULL;
				CR();
				write2HWserie((CONV)"Latitude: ");		// Displays data in suitable format
				writeData2HWserie(latitudeDeg);
				write2HWserie((CONV)" ");
				writeData2HWserie(N_S);
				write2HWserie((CONV)"   ");
				write2HWserie((CONV)"Longitude: ");
				writeData2HWserie(longitudeDeg);
				write2HWserie((CONV)" ");
				writeData2HWserie(E_W);
				write2HWserie((CONV)"   ");
				write2HWserie((CONV)"Altitude: ");
				writeData2HWserie(altitude);
				write2HWserie((CONV)" m   ");
				write2HWserie((CONV)"Velocity: ");
				writeData2HWserie(speedGround);
				write2HWserie((CONV)" km/h  ");
				input=0;											// Read switches SW2 and SW1
				input=((unsigned char)SW2<<1) | (unsigned char)SW1;
				switch(input)
				{
					case 0:
						latitudeDeg[2]=214;
						writeLCDmsg((const far rom char *)"Lat:");
						GotoXY(0,4);
						for(i=0;i<4;i++) PutLCD(latitudeDeg[i]);
						GotoXY(1,0);
						for(i=4;i<strlen(latitudeDeg);i++) PutLCD(latitudeDeg[i]);
						GotoXY(1,7);
						PutLCD(N_S[0]);
						break;
					case 1:
						longitudeDeg[3]=214;
						writeLCDmsg((const far rom char *)"Lo:");
						GotoXY(0,3);
						for(i=0;i<5;i++) PutLCD(longitudeDeg[i]);
						GotoXY(1,0);
						for(i=5;i<strlen(longitudeDeg);i++) PutLCD(longitudeDeg[i]);
						GotoXY(1,7);
						PutLCD(E_W[0]);
						break;
					case 2:
						writeLCDmsg((const far rom char *)" Alt: ");
						GotoXY(0,6);
						for(i=0;i<2;i++) PutLCD(altitude[i]);
						GotoXY(1,0);
						for(i=2;i<strlen(altitude);i++) PutLCD(altitude[i]);
						GotoXY(1,4);
						writeLCDmsgRaw((const far rom char *)" m  ");
						break;
					case 3:
						writeLCDmsg((const far rom char *)"Vel: ");
						GotoXY(0,5);
						for(i=0;i<3;i++) PutLCD(speedGround[i]);
						GotoXY(1,0);
						for(i=3;i<strlen(speedGround);i++) PutLCD(speedGround[i]);
						GotoXY(1,4);
						writeLCDmsgRaw((const far rom char *)"km/h");
						break;
				}
			}
			else 	
			{
				susCharX(latitude,9);
				susCharX(N_S,1);
				susCharX(longitude,10);
				susCharX(E_W,1);
				susCharX(altitude,6);
				CR();
				write2HWserie((CONV)"Latitude: ");
				writeData2HWserie(latitude);
				write2HWserie((CONV)" ");
				writeData2HWserie(N_S);
				write2HWserie((CONV)"   ");
				write2HWserie((CONV)"Longitude: ");
				writeData2HWserie(longitude);
				write2HWserie((CONV)" ");
				writeData2HWserie(E_W);
				write2HWserie((CONV)"   ");
				write2HWserie((CONV)"Altitude: ");
				writeData2HWserie(altitude);
				write2HWserie((CONV)" m   ");
				write2HWserie((CONV)"Velocity: ");
				writeData2HWserie(speedGround);
				write2HWserie((CONV)" km/h  ");
				input=0;
				input=((unsigned char)SW2<<1) | (unsigned char)SW1;
				switch(input)
				{
					case 0:
						writeLCDmsg((const far rom char *)"Lat: unavailable");
						break;
					case 1:
						writeLCDmsg((const far rom char *)"Lon: unavailable");
						break;
					case 2:
						writeLCDmsg((const far rom char *)"Alt: unavailable");
						break;
					case 3:
						writeLCDmsg((const far rom char *)"Vel: unavailable");
						break;
				}
			}		
		}
		else 	
		{
			CR();
			write2HWserie((CONV)"Unable to access to GPS information!");
			writeLCDmsg((const far rom char *)"GPS unavailable!");
		}
	}
}

void writeLCDmsg(const far rom char * message)
{
	unsigned char i;
	ClearScreen();							// Clear screen
	GotoXY(0,0);							
	for(i=0; i<8; i++) PutLCD(message[i]);	// Writes bytes to serial port
	GotoXY(1,0);
	for(i=8; i<strlenpgm(message);i++) PutLCD(message[i]);
}

void writeLCDmsgRaw(const far rom char * mess)
{
	unsigned char i;
	for(i=0; i<strlenpgm(mess); i++) PutLCD(mess[i]);	
}

void writeData2LCD(char * data)
{
	unsigned char i;
	ClearScreen();							
	GotoXY(0,0);							
	for(i=0; i<8; i++) PutLCD(data[i]);		
	GotoXY(1,0);
	for(i=8; i<strlen(data);i++) PutLCD(data[i]);
}

void susCharX(char * splitArray, int nChars)
{
	int i;
	for(i=0; i<nChars;i++ ) splitArray[i]='X';
	splitArray[i]=0;
}

void flush(void)
{
	int initIndex;
	for(initIndex=0;initIndex<INCOMING_SIZE_1;initIndex++) serialIncomingData_1[initIndex]=0;	//Flush input buffers
	for(initIndex=0;initIndex<INCOMING_SIZE_2;initIndex++) serialIncomingData_2[initIndex]=0;	//
}

int checksum(char * serialData, char * checksumData)		// Check if data is reliable
{
	unsigned char i;
	char checksum=0;
	char checksumArray[3];
	for(i=1; i<(strlen(serialData)-1);i++)
	{
		checksum=checksum^serialData[i];
	}
	convert2hex(checksumArray,checksum);
	if(checksumArray[0]==checksumData[0] && checksumArray[1]==checksumData[1])
	{ 
		 return OK;
	}
	return ERROR;
}

void writeData2HWserie(char *data)
{
	int i;
	for(i=0; i<strlen(data); i++)
	{
		while(BusyUSART());				// Wait until serial port is available
		putcUSART(data[i]);
	}
}

void write2HWserie(const far rom char* message)
{
	int i;
	for(i=0; i<strlenpgm(message); i++)
	{
		while(BusyUSART());
		putcUSART(message[i]);
	}
}

void convert2hex(char* checksumArray, unsigned char data)
{
 	checksumArray[0] = ((data>>4) & 0x0F) + '0';
  	checksumArray[1] = ((data)    & 0x0F) + '0';
  	checksumArray[2] = 0;  
  	if(checksumArray[0] > '9') checksumArray[0] += 7;		// Char to hexadecinal ascii character
  	if(checksumArray[1] > '9') checksumArray[1] += 7;
}

void initConfig(void)
{
	int initIndex=0;
	init_bolt();
	activa_serial();
	OpenUART();		
	InitLCD();
	flush();
	sendCommand2GPS((CONV)TEST);			//Test Packet
	sendCommand2GPS((CONV)FREQUENCY);		//5 s update rate
	sendCommand2GPS((CONV)GGA_VTG);			//GGA and VTG
	sendCommand2GPS((CONV)DATUM_1);			//Datum for Mexico
	sendCommand2GPS((CONV)HOT);				//Hot restart
}

void sendCommand2GPS(const far rom char * command)
{
	int i;
	for(i=0; i<strlenpgm(command);i++)
	{	
		putcUART(command[i]);
		delay_ms(2);
	}
	putcUART(13);	//<CR>
	delay_ms(2);
	putcUART(10);	//<LF>
	delay_ms(1000);
}

void DelayTXBitUART(void)
{
   unsigned char delay_time = (((((2*CLK_factor)/(4*BAUD_factor))+1)/2)-12)/10;
   Delay10TCYx(delay_time);
   return;
}

void DelayRXHalfBitUART(void)
{
   unsigned char delay_time = (((((2*CLK_factor)/(8*BAUD_factor))+1)/2)-9)/10;
   Delay10TCYx(delay_time);
   return;
}

void DelayRXBitUART(void)
{
   unsigned char delay_time = (((((2*CLK_factor)/(4*BAUD_factor))+1)/2)-14)/10;
   Delay10TCYx(delay_time);
   return;
}



